      program testrmp
      implicit none
#include <define.h>
#include <rmp_mpif.h>
#include <rmp_rank.h>
#include <rmp_index.h>
      integer imf, jmf, kmf, km, kb, npes, ncol, nrow, lb, im, jm
      parameter(imf=1028,jmf=1028,lb=2,km=72,kb=1,kmf=km)
      parameter(npes=npes_,ncol=ncol_,nrow=nrow_)
      parameter(im =(imf-lb*2-1)/ncol+1+lb*2)
      parameter(jm =(jmf-lb*2-1)/nrow+1+lb*2)
c
      real(kind=8) part1(im,jm,km)
      real(kind=8) part2(im,jm,km)
      real(kind=8) part3(im,jm,km)
      real(kind=8) wstart, wstop
      integer DispPackNS(lb*km), DispUnpackNS(lb*km)
      integer DispPackEW(jm*km), DispUnpackEW(jm*km)
#ifdef OUTPUT
      integer GlobalSizes_3d(3), LocalSizesWithHalo_3d(3), LocalSizesSansHalo_3d(3)
      integer LocalStarts_3d(3), GlobalStartsWrite_3d(3)
      integer memSubDomain_3d_r8, FileSubDomain_3d_r8
      integer (kind=MPI_OFFSET_KIND) :: disp=0
      integer (kind=MPI_OFFSET_KIND) :: offset=0
      integer info, unit
#endif
      integer i,j,k,status,h,mk,ibc,np,ierr

! start MPI
! ---------
      call MPI_Init(ierr)
      call MPI_Comm_size( MPI_COMM_WORLD, nsize , ierr )
      call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
      mynum = myrank + 1

      call MPI_Comm_set_errhandler(MPI_COMM_WORLD, MPI_ERRORS_RETURN, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Comm_set_errhandler ',ierr
 
! check the cosistency of decomposition
! -------------------------------------
      np=ncol * nrow
      if( np.ne.npes ) then
        print *,' Error: inconsistent decomposition '
        print *,' npes should be equal to ncol * nrow '
        print *,' npes = ',npes
        print *,' ncol = ',ncol
        stop
      else
!        print *,' my full dimension imf jmf and lb are ',imf,jmf,lb
!        print *,' my decomposition ncol nrow are ',ncol,nrow
!        print *,' my partial dimension im jm are ',im,jm
      endif
 
! start decomposition
! -------------------
      call rmp_equdiv(imf,lb,ncol,myistr,myilen,npes, ncol)
!      print *,' PE',myrank,' istr ilen ',myistr(mynum),myilen(mynum)
 
      call rmp_equdiv(jmf,lb,nrow,myjstr,myjlen,npes,-ncol)
!      print *,' PE',myrank,' jstr jlen ',myjstr(mynum),myjlen(mynum)
 
      call rmp_equdiv(kmf,kb,ncol,mykstr,myklen,npes, ncol)
!      print *,' PE',myrank,' kstr klen ',mykstr(mynum),myklen(mynum)
 
      call rmp_equdiv(imf,lb,nrow,mymstr,mymlen,npes,-ncol)
!      print *,' PE',myrank,' mstr mlen ',mymstr(mynum),mymlen(mynum)
 
      call rmp_equdiv(jmf,lb,ncol,mynstr,mynlen,npes, ncol)
!      print *,' PE',myrank,' nstr nlen ',mynstr(mynum),mynlen(mynum)
 
      call rmp_equdiv(kmf,kb,nrow,mylstr,myllen,npes,-ncol)
!      print *,' PE',myrank,' lstr llen ',mylstr(mynum),myllen(mynum)
 
      ibc=0
      call rmp_neighbor(npes,ncol,mynebtt,mynebbb,mynebll,mynebrr,ibc)
!      print *,' PE',myrank,'   neighor top ',mynebtt(mynum)
!      print *,' PE',myrank,'   neighor bot ',mynebbb(mynum)
!      print *,' PE',myrank,'   neighor lft ',mynebll(mynum)
!      print *,' PE',myrank,'   neighor rgh ',mynebrr(mynum)

#ifdef OUTPUT
      call MPI_Info_create(info, ierr)
      call MPI_Info_set(info, "direct_read", "false", ierr)
      call MPI_Info_set(info, "romio_cb_read", "enable", ierr)
      call MPI_Info_set(info, "direct_write", "false", ierr)
      call MPI_Info_set(info, "romio_cb_write", "enable", ierr)

! Now setup 3D
! ------------
      GlobalSizes_3d(1) = im*ncol
      GlobalSizes_3d(2) = jm*nrow
      GlobalSizes_3d(3) = kmf
      LocalSizesWithHalo_3d(1) = im        ; LocalSizesWithHalo_3d(2) = jm        ; LocalSizesWithHalo_3d(3) = km
      LocalSizesSansHalo_3d(1) = im        ; LocalSizesSansHalo_3d(2) = jm        ; LocalSizesSansHalo_3d(3) = km
      LocalStarts_3d(1)        = 0         ; LocalStarts_3d(2)        = 0         ; LocalStarts_3d(3)        = 0
      GlobalStartsWrite_3d(1)  = im*mod(myrank,ncol)
      GlobalStartsWrite_3d(2)  = jm*(myrank/ncol)
      GlobalStartsWrite_3d(3)  = 0

! Customize counts and offsets for boundary processes
! ---------------------------------------------------
      if(myistr(myrank+1) == 1) then
         LocalSizesSansHalo_3d(1) = im-(1*lb)
         LocalStarts_3d(1)        = 0
         GlobalStartsWrite_3d(1)  = myistr(myrank+1)-1
      endif
      if(myjstr(myrank+1) == 1) then
         LocalSizesSansHalo_3d(2) = jm-(1*lb)
         LocalStarts_3d(2)        = 0
         GlobalStartsWrite_3d(2)       = myjstr(myrank+1)-1
      endif
      if(myistr(myrank+1)+myilen(myrank+1)-1 == imf) then
         LocalSizesSansHalo_3d(1) = im-(1*lb)
      endif
      if(myjstr(myrank+1)+myjlen(myrank+1)-1 == jmf) then
         LocalSizesSansHalo_3d(2) = jm-(1*lb)
      endif

!      print*,myrank,' GlobalSizes_3d ',GlobalSizes_3d
!      print*,myrank,' GlobalStartsWrite_3d ',GlobalStartsWrite_3d

! Following http://user.cscs.ch/software_and_programming_environment/compilers_and_programming/rosa_cray_xt5/lustre_io_optimization/index.html
! Create mapping between subdomains with halos to subdomains sans halo
! --------------------------------------------------------------------
      call MPI_Type_create_subarray(3, LocalSizesWithHalo_3d, LocalSizesWithHalo_3d, LocalStarts_3d,
     .                              MPI_ORDER_FORTRAN, MPI_REAL8, memSubDomain_3d_r8, ierr)
      call MPI_Type_commit(memSubDomain_3d_r8, ierr)

! Create mapping between subdomains sans halo and Global domain
! -------------------------------------------------------------
      call MPI_Type_create_subarray(3, GlobalSizes_3d, LocalSizesWithHalo_3d, GlobalStartsWrite_3d,
     .                              MPI_ORDER_FORTRAN, MPI_REAL8, FileSubDomain_3d_r8, ierr)
      call MPI_Type_commit(FileSubDomain_3d_r8, ierr)
#endif

! Define the North halo region
! ----------------------------
      i=lb+1; mk=1
      do k=1,kmf
         h=0
         do j=myjlen(mynum)-lb*2+1,myjlen(mynum)-lb
            DispPackNS(mk+h)   = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(myrank == 0 .or. myrank == 4) print*,myrank,': DispPackN = ',DispPackNS(mk+h)
            h=h+1
         enddo
         h=0
      do j=myjlen(mynum)-lb+1,myjlen(mynum)
            DispUnpackNS(mk+h) = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(myrank == 0 .or. myrank == 4) print*,myrank,': DispUnpackN = ',DispUnpackNS(mk+h)
            h=h+1
         enddo
         mk=mk+lb
      enddo

      call MPI_Type_create_indexed_block(lb*kmf, myilen(mynum)-lb*2, DispPackNS  , MPI_REAL8, NorthPack  , ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block1 ',ierr
      call MPI_Type_commit(NorthPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of NorthPack',ierr

      call MPI_Type_create_indexed_block(lb*kmf, myilen(mynum)-lb*2, DispUnpackNS, MPI_REAL8, NorthUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block2 ',ierr
      call MPI_Type_commit(NorthUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of NorthUnpack',ierr
 
! Define the South halo region
! ----------------------------
      i=lb+1; mk=1
      do k=1,kmf
         h=0
         do j=lb+1,2*lb
            DispPackNS(mk+h)   = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(myrank == 0 .or. myrank == 1) print*,myrank,': DispPackS = ',k,DispPackNS(mk+h),mk+h
!            print*,myrank,': DispPackS = ',DispPackNS(mk+h)
            h=h+1
         enddo
         h=0
         do j=1,lb
            DispUnpackNS(mk+h) = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(myrank == 0 .or. myrank == 1) print*,myrank,': DispUnpackS = ',k,DispUnpackNS(mk+h),mk+h
!            print*,myrank,': DispUnpackS = ',DispUnpackNS(mk+h)
            h=h+1
         enddo
         mk=mk+lb
      enddo

      call MPI_Type_create_indexed_block(lb*kmf, myilen(mynum)-lb*2, DispPackNS  , MPI_REAL8, SouthPack  , ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block3 ',ierr
      call MPI_Type_commit(SouthPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of SouthPack',ierr

      call MPI_Type_create_indexed_block(lb*kmf, myilen(mynum)-lb*2, DispUnpackNS, MPI_REAL8, SouthUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block4 ',ierr
      call MPI_Type_commit(SouthUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of SouthUnpack',ierr

! Define the West halo region
! ---------------------------
      mk=1
      do k=1,kmf
         i=lb+1
         h=0
         do j=1,myjlen(mynum)
            DispPackEW(mk+h)   = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(k==1 .and. (myrank == 0 .or. myrank == 1)) print*,myrank,': DispPackW = ',DispPackEW(mk+h)
!            if(myrank == 0 .or. myrank == 1) print*,myrank,': DispPackW = ',k,DispPackEW(mk+h),mk+h
            h=h+1
         enddo
         i=1
         h=0
         do j=1,myjlen(mynum)
            DispUnPackEW(mk+h) = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(k==1 .and. (myrank == 0 .or. myrank == 1)) print*,myrank,': DispUnpackW = ',DispUnpackEW(mk+h)
!            if(myrank == 0 .or. myrank == 1) print*,myrank,': DispUnpackW = ',k,DispUnpackEW(mk+h),mk+h
            h=h+1
         enddo
         mk=mk+myjlen(mynum)
      enddo

      call MPI_Type_create_indexed_block(myjlen(mynum)*kmf, lb, DispPackEW, MPI_REAL8, WestPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block4 ',ierr
      call MPI_Type_commit(WestPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of WestPack',ierr
      call MPI_Type_create_indexed_block(myjlen(mynum)*kmf, lb, DispUnpackEW, MPI_REAL8, WestUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block5 ',ierr
      call MPI_Type_commit(WestUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of WestUnpack ',ierr

! Define the East halo region
! ---------------------------
      mk=1
      do k=1,kmf
         i=myilen(mynum)-2*lb+1
         h=0
         do j=1,myjlen(mynum)
            DispPackEW(mk+h)   = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(k==1 .and. (myrank == 0 .or. myrank == 1)) print*,myrank,': DispPackE = ',DispPackEW(mk+h)
            h=h+1
         enddo
         i=myilen(mynum)-lb+1
         h=0
         do j=1,myjlen(mynum)
            DispUnPackEW(mk+h) = (i+((j-1)*myilen(mynum))+((k-1)*myilen(mynum)*myjlen(mynum)))-1
!            if(k==1 .and. (myrank == 0 .or. myrank == 1)) print*,myrank,': DispUnpackE = ',DispUnpackEW(mk+h)
            h=h+1
         enddo
         mk=mk+myjlen(mynum)
      enddo

      call MPI_Type_create_indexed_block(myjlen(mynum)*kmf, lb, DispPackEW, MPI_REAL8, EastPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block6 ',ierr
      call MPI_Type_commit(EastPack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of EastPack',ierr
      call MPI_Type_create_indexed_block(myjlen(mynum)*kmf, lb, DispUnpackEW, MPI_REAL8, EastUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_create_indexed_block7 ',ierr
      call MPI_Type_commit(EastUnpack, ierr)
      if(ierr /= MPI_SUCCESS) print*,': Failure during MPI_Type_commit of EastUnpack',ierr
 
! fill with arbitrary data
! ------------------------
      do k=1,kmf
        do j=1,jm
          do i=1,im
            part1(i,j,k)=dble(myrank)
          enddo
        enddo
      enddo

! DSK set halo corners to a signalling NaN
! leave this here in case ajmone wants to figure out why physics is using the halo corners
! ----------------------------------------------------------------------------------------
      do k=1,kmf
         part1(1,1,k)=Z'7FFC000000000000'
         part1(2,1,k)=Z'7FFC000000000000'
         part1(1,2,k)=Z'7FFC000000000000'
         part1(2,2,k)=Z'7FFC000000000000'

         part1(1,myjlen(mynum)-lb+1,k)=Z'7FFC000000000000'
         part1(2,myjlen(mynum)-lb+1,k)=Z'7FFC000000000000'
         part1(1,myjlen(mynum)     ,k)=Z'7FFC000000000000'
         part1(2,myjlen(mynum)     ,k)=Z'7FFC000000000000'

         part1(myilen(mynum)-lb+1,1,k)=Z'7FFC000000000000'
         part1(myilen(mynum)     ,1,k)=Z'7FFC000000000000'
         part1(myilen(mynum)-lb+1,2,k)=Z'7FFC000000000000'
         part1(myilen(mynum)     ,2,k)=Z'7FFC000000000000'

         part1(myilen(mynum)-lb+1,myjlen(mynum)-lb+1,k)=Z'7FFC000000000000'
         part1(myilen(mynum)     ,myjlen(mynum)-lb+1,k)=Z'7FFC000000000000'
         part1(myilen(mynum)-lb+1,myjlen(mynum)     ,k)=Z'7FFC000000000000'
         part1(myilen(mynum)     ,myjlen(mynum)     ,k)=Z'7FFC000000000000'
      enddo

      part2=part1

! call update halo
! ----------------
      part3=part2
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo_DerType(part1,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo_DerType ',(wstop-wstart)*1.e6,' us'

      part3=part1
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo(part2,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo ',(wstop-wstart)*1.e6,' us'

      part3=part2
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo_DerType(part1,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo_DerType ',(wstop-wstart)*1.e6,' us'

      part3=part1
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo(part2,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo ',(wstop-wstart)*1.e6,' us'

      part3=part2
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo_DerType(part1,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo_DerType ',(wstop-wstart)*1.e6,' us'

      part3=part1
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstart = MPI_Wtime()
      call rmp_updatehalo(part2,im,jm,kmf,lb)
      call MPI_Barrier(MPI_COMM_WORLD, status)
      wstop = MPI_Wtime()
      if( myrank.eq.master ) print *,'updatehalo ',(wstop-wstart)*1.e6,' us'

#ifdef OUTPUT
! make sure file exists
! ---------------------
      if (myrank == 0) then
         call MPI_File_open(MPI_COMM_SELF, '/nobackup/dkokron/GCE_halo', MPI_MODE_WRONLY+MPI_MODE_CREATE, MPI_INFO_NULL, unit, ierr)
         if(ierr /= MPI_SUCCESS) print*,'Failure to create file dynamics restart',ierr
         call MPI_File_close(unit, ierr)
         if(ierr /= MPI_SUCCESS) print*,'Failure to close file dynamics restart',ierr
      endif
      call MPI_Barrier(MPI_COMM_WORLD, ierr)
      call MPI_File_open(MPI_COMM_WORLD, '/nobackup/dkokron/GCE_halo', MPI_MODE_WRONLY+MPI_MODE_CREATE, MPI_INFO_NULL, unit, ierr)
      if(ierr /= MPI_SUCCESS) print*,'Failure to open ',ierr

      call MPI_File_get_position(unit, offset, ierr)
      if(ierr /= MPI_SUCCESS) print*,'Failure during 3d get_position ',ierr
      call MPI_File_get_byte_offset(unit, offset, disp, ierr)
      if(ierr /= MPI_SUCCESS) print*,'Failure during 3d get_byte_offset ',ierr
      call MPI_File_set_view(unit, disp, MPI_DOUBLE_PRECISION, FileSubDomain_3d_r8, "native", MPI_INFO_NULL, ierr)
      if(ierr /= MPI_SUCCESS) print*,'Failure during 3D set_view ',ierr

! Write 3D fields
! ---------------
      call MPI_File_write_all(unit, part1, im*jm*km, MPI_REAL8, MPI_STATUS_IGNORE, ierr)
      call MPI_Barrier(MPI_COMM_WORLD, ierr)
      call MPI_File_close(unit, ierr)
#endif

!      do k=1,kmf
!        do j=1,jm
!          do i=1,im
!            if(isnan(part1(i,j,k))) print*,'Found NaN at ',i,j,k
!          enddo
!        enddo
!      enddo

! end MPI
! -------
      call MPI_Type_free(NorthPack, ierr)
      call MPI_Type_free(NorthUnpack, ierr)
      call MPI_Type_free(SouthPack, ierr)
      call MPI_Type_free(SouthUnpack, ierr)
      call MPI_Type_free(WestPack, ierr)
      call MPI_Type_free(WestUnpack, ierr)
      call MPI_Type_free(EastPack, ierr)
      call MPI_Type_free(EastUnpack, ierr)
      call MPI_Comm_free(comm_row, ierr)
      call MPI_Comm_free(comm_col, ierr)
!      call MPI_Comm_free(comm_cart, ierr)
#ifdef OUTPUT
      call MPI_Info_free(Info,ierr)
      call MPI_Type_free(memSubDomain_3d_r8, ierr)
      call MPI_Type_free(FileSubDomain_3d_r8, ierr)
#endif

      call MPI_Finalize(ierr)

      end program

